!!  Poisson equation solver
!!  div(grad u) = f(x,y)
!!  f = x * exp(y)
!!  Exact solution is u = x * exp(y)

      Program PoissonSolver
      implicit NONE
      include 'PARAM'
      
      real*8 u(N_x,N_y),u_iter(N_x,N_y),denom,factor,x,y,res, &
             x_left,x_right,y_left,y_right,dx,dy,res_max,f

      integer i,j,it

      character*8  SOLTIM
      character*10 SOLDAT

      x_left  = 0.d0
      x_right = 1.d0
      y_left  = 0.d0
      y_right = 1.d0
      dx = (x_right-x_left)/real(N_x - 1)
      dy = (y_right-y_left)/real(N_y - 1)

100   format(1x,4(2x,e12.5))
200   format(1x,I4,e12.5)
614   format('# Date: ',A10,'   Time: ',A8)

!! boundary conditions are set 
      do 10 j = 1,N_y
        y = dy * real(j)
        u     (  1,j) = 0.d0
        u     (N_x,j) = x_right * exp(y)
        u_iter(  1,j) = 0.d0
        u_iter(N_x,j) = x_right * exp(y)
10    continue

      do 20 i = 1,N_x
        x = dx * real(i)
        u     (i,  1) = x
        u     (i,N_y) = x * exp(y_right)
        u_iter(i,  1) = x
        u_iter(i,N_y) = x * exp(y_right)
20    continue

!! intial values are given for interior points
      do 30 j = 2,N_y-1
      do 30 i = 2,N_x-1
        u(i,j) = 0.d25 * (u(i,0)+u(i,N_y)+u(0,j)+u(N_x,j))
30    continue

      denom = 2.d0 / dx**2 + 2.d0 / dy**2
      factor = 1 / denom

      open(19,file='history.dat')
      write(19,*) 'Variables = Iter, Res'
      call GetDate(SOLDAT,SOLTIM)
      write(19,614) SOLDAT,SOLTIM

!$omp parallel default(none)        &
!$omp shared(u,u_iter)              &
!$omp private(x,y,i,j,res,it)       &
!$omp firstprivate(x_left,y_left,dx,dy,factor) &
!$omp reduction(max: res_max) 

!! start iterations
      do 40 it = 1,iter_max
!$omp do
        do 50 j = 2,N_y - 1
        do 50 i = 2,N_x - 1
          x = x_left + dx * real(i)
          y = y_left + dy * real(j)
          u_iter(i,j) = (u(i+1,j) + u(i-1,j))/dx**2 + &
                        (u(i,j+1) + u(i,j-1))/dy**2 - f(x,y)
          u_iter(i,j) =  u_iter(i,j) * factor
50    continue

          res_max = 0.d0
!$omp do
          do 60 j = 2,N_y-1
          do 60 i = 2,N_x-1
        if(mod(it,100) == 0)then
            res = abs(u(i,j) - u_iter(i,j))
            if(res_max < res) res_max = res
        endif
          u(i,j) = u_iter(i,j)
60      continue
40    continue
!$omp end parallel

      write(19,200) iter_max, res_max
      call GetDate(SOLDAT,SOLTIM)
      write(19,614) SOLDAT,SOLTIM
      close(19)

!! write part
      if(key_print) then
        open(19,file='u.dat')
        write(19,*) 'Variables = x, y, u, f'
        write(19,*) 'Zone i = ',N_x,' j = ',N_y
        call GetDate(SOLDAT,SOLTIM)
        write(19,614) SOLDAT,SOLTIM
        do 70 j = 1,N_y,1
        do 70 i = 1,N_x,1
          x = dx*real(i)
          y = dy*real(j)
          write(19,100) x,y,u(i,j),f(x,y)
70      continue
        close(19)
      endif

      stop
      end

      function f(x,y)
        implicit NONE
        real*8 f
        real*8 x,y
          f = x * exp(y)
      end function f

      subroutine GetDate(SDATE, STIME) 
        implicit NONE
        character*10 SDATE
        character*8  STIME
        integer date_time(8)
        integer*2 year,month,day,hrs,mins,secs

          call date_and_time(values=date_time)
          year  = date_time(1)
          month = date_time(2)
          day   = date_time(3)
          hrs   = date_time(5)
          mins  = date_time(6)
          secs  = date_time(7)
          write(sdate,100) day, month, year
          write(stime,200) hrs, mins, secs
100   FORMAT(I2.2,'.',I2.2,'.',I4.4)
200   FORMAT(I2.2,':',I2.2,':',I2.2)
      return
      end
